pub const CFGNAME: &str = "config.ron";
use eyre::Context;
use ron::de::from_reader;
use ron::ser::{to_string_pretty, PrettyConfig};
use serde::{Deserialize, Serialize};
use std::env;
use std::io::Write;
use std::{fs::File, io::ErrorKind, net::SocketAddr, str::FromStr};
use tracing::{error, event, Level};

#[derive(Debug, Clone, Default, Serialize, Deserialize)]
pub struct Conf {
    pub core: Core,
    pub logging: Logging,
    pub web: Web,
    pub api: Api,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Core {
    pub socketaddr: SocketAddr,
    pub threads: Option<usize>,
    pub cache_thread_pool: Option<bool>,
}

impl Default for Core {
    fn default() -> Self {
        Self {
            socketaddr: SocketAddr::from_str("127.0.0.1:8086").unwrap(),
            threads: Default::default(),
            cache_thread_pool: Some(false),
        }
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Logging {
    pub log_level: String,
    pub custom_filter: Option<String>,
}

impl Default for Logging {
    fn default() -> Self {
        Self {
            log_level: "info".into(),
            custom_filter: Default::default(),
        }
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Web {
    pub template_dir: String,
    pub asset_url: String,
    pub serve_assets: Option<String>,
}

impl Default for Web {
    fn default() -> Self {
        Self {
            template_dir: "templates".into(),
            asset_url: Default::default(),
            serve_assets: Some("./static".into()),
        }
    }
}

#[derive(Debug, Default, Clone, Serialize, Deserialize)]
pub struct Api {
    pub steam_key: String,
    pub stats_url: String,
}

/// Loads the program configuration from `config.ini` beside the executable.
///
/// Returns an error if the configuration file:
/// * Cannot be parsed
/// * Does not exist, and cannot be created.
/// * Cannot be accessed
pub fn get_settings() -> Result<Conf, eyre::Report> {
    let _span = tracing::span!(tracing::Level::TRACE, "Loading Configuration").entered();

    let mut path = env::current_exe()?;
    path.pop();
    path.push(CFGNAME);

    let configuration;
    // Check if the cfg file exists, and can be written to.
    event!(Level::DEBUG, "Reading…");
    match File::open(&path) {
        Ok(f) => {
            configuration =
                from_reader(f).wrap_err(format!("Deserializing '{}'", path.display()))?;
            event!(Level::DEBUG, "Read.")
        }
        Err(e) => match e.kind() {
            ErrorKind::NotFound => {
                event!(Level::WARN, "Configuration not found, writing…");
                let mut file = File::create(&path)?;
                let cfg = Conf::default();
                let pretty = PrettyConfig::new()
                    .depth_limit(2)
                    .separate_tuple_members(true)
                    .enumerate_arrays(true);
                file.write_all(to_string_pretty(&cfg, pretty)?.as_bytes())?;
                configuration = cfg;
                event!(
                    Level::WARN,
                    "Wrote default configuration to: \"{}\"",
                    &path.to_string_lossy()
                );
            }
            _ => {
                return Err(e).wrap_err_with(|| {
                    format!("Unable to access cfg file {}", &path.to_string_lossy())
                })
            }
        },
    };
    Ok(configuration)
}

/// Initializes logging, and attemps to load the config file.
pub fn get_config_init_logging() -> Result<Conf, eyre::Error> {
    use tracing_subscriber::filter::{EnvFilter, LevelFilter};

    // Logging and CFG
    #[cfg(target_os = "windows")]
    {
        ansi_term::enable_ansi_support()
            .map_err(|e| eyre::Report::msg(format!("Cannot enable ANSI color on Windows: {}", e)))?
    };
    // TRACE level events will be output here.
    let builder = tracing_subscriber::fmt()
        .with_env_filter(EnvFilter::default().add_directive(LevelFilter::TRACE.into()))
        .with_filter_reloading();
    let filter_handle = builder.reload_handle();
    builder.init();
    // Load CFG, this should be done after logging is initialized.
    let cfg = get_settings()?;
    let mut new_filter = EnvFilter::default().add_directive(cfg.logging.log_level.parse()?);
    let custom_filter = cfg
        .logging
        .custom_filter
        .clone()
        .unwrap_or(String::default());
    if !custom_filter.is_empty() {
        match tracing_subscriber::filter::Directive::from_str(custom_filter.as_str()) {
            Ok(v) => new_filter = new_filter.add_directive(v),
            Err(e) => {
                error!("Unable to parse custom filter: '{}' - {}", custom_filter, e);
            }
        }
    }
    // Switch to user-defined logging level.
    filter_handle.modify(|layer| *layer = new_filter)?;
    Ok(cfg)
}
