use crate::{steamapi::Player, CONFIGURATION};
use chrono::Duration;
use eyre::Error;
use moka::sync::Cache;
use serde::{Deserialize, Serialize, Serializer};
use tracing::{error, info};

// Moka cache
// This prevents having to hit the Stats API as often, and reduces second
// request latency.
lazy_static! {
    pub static ref STATS_CACHE: Cache<u64, Stats> = {
        info!("Initializing cache");
        let pool = CONFIGURATION.core.cache_thread_pool.unwrap_or(false);
        Cache::builder()
            .time_to_live(std::time::Duration::from_secs(60 * 5)) // Five minutes.
            .max_capacity(256)
            .thread_pool_enabled(pool)
            .build()
    };
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Stats {
    pub steamid64: u64,
    #[serde(skip_serializing)]
    pub deaths: u64,
    #[serde(skip_serializing)]
    pub shots: u64,
    #[serde(skip_serializing)]
    pub suicides: u64,
    #[serde(skip_serializing)]
    pub loot: u64,
    #[serde(skip_serializing)]
    pub rounds: u64,
    #[serde(serialize_with = "serialize_duration")]
    pub playtime: i64,
    #[serde(skip_serializing)]
    pub steps: u64,
    #[serde(rename = "Frags")]
    pub frags: Frags,
    #[serde(rename = "Wins")]
    pub wins: Wins,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Frags {
    #[serde(rename = "Dfrags")]
    pub d_frags: u64,
    #[serde(rename = "Tfrags")]
    pub t_frags: u64,
    #[serde(rename = "Ifrags")]
    pub i_frags: u64,
    pub frags: u64,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Wins {
    #[serde(rename = "Twins")]
    pub t_wins: u64,
    #[serde(rename = "Dwins")]
    pub d_wins: u64,
    #[serde(rename = "Iwins")]
    pub i_wins: u64,
}

impl Default for Stats {
    fn default() -> Self {
        Self {
            steamid64: Default::default(),
            deaths: Default::default(),
            shots: Default::default(),
            suicides: Default::default(),
            loot: Default::default(),
            rounds: Default::default(),
            playtime: Default::default(), // 356521
            steps: Default::default(),
            frags: Frags {
                d_frags: Default::default(),
                t_frags: Default::default(),
                i_frags: Default::default(),
                frags: Default::default(),
            },
            wins: Wins {
                t_wins: Default::default(),
                d_wins: Default::default(),
                i_wins: Default::default(),
            },
        }
    }
}

lazy_static! {
    static ref CLIENT: reqwest::Client = {
        reqwest::Client::builder()
            .build()
            .expect("Trying to initalize Reqwest client for SteamAPI")
    };
}

#[tracing::instrument(level = "info")]
pub async fn get_player_stats(player: &Player) -> Result<Stats, Error> {
    // Check if we have it in cache already.
    match STATS_CACHE.get(&player.steamid) {
        Some(v) => return Ok(v),
        None => {}
    }
    // Otherwise do the work...
    let client = (*CLIENT).clone();
    let api_url = &CONFIGURATION.api.stats_url;
    let data = client
        .get(format!("{api_url}/{}", player.steamid))
        .send()
        .await?
        .text()
        .await?;
    let stats: Stats = serde_json::from_str(&data).map_err(|e| {
        error!("Deserializing stats: '{}'", e);
        e
    })?;
    STATS_CACHE.insert(player.steamid, stats.clone());
    Ok(stats)
}

// Convert to a human redable duration, not round-trip-able
fn serialize_duration<S>(secs: &i64, s: S) -> Result<S::Ok, S::Error>
where
    S: Serializer,
{
    let dur = Duration::seconds(*secs);
    let mut buffer = String::new();
    // Prepare for some serious ugly.
    let days = dur.num_days();
    let hours = dur.num_hours() - (days * 24);
    let minutes = dur.num_minutes() - (dur.num_hours() * 60);
    if days > 0 {
        buffer.push_str(format!("{} days, ", days).as_str());
    };
    if hours > 0 {
        buffer.push_str(format!("{} hours, ", hours).as_str());
    };
    buffer.push_str(format!("{} minutes", minutes).as_str());

    s.serialize_str(&buffer)
}
