use std::time::Duration;

use moka::sync::Cache;
use reqwest::Client;
use serde::{Deserialize, Serialize};
use serde_json::Value;
use tracing::{debug, info, warn};

use crate::CONFIGURATION;

const API_URL: &str = "https://api.steampowered.com/ISteamUser/GetPlayerSummaries/v2";
//                                        ↓ valid alternative to Akamai.
const AVATAR_URL: &str = "https://avatars.cloudflare.steamstatic.com";
const POSTFIX: &str = "_full.jpg";
const DEFAULT_HASH: &str = "fef49e7fa7e1997310d705b2a6158ff8dc1cdfeb";

// Moka cache
// This prevents having to hit the Steam API as often, and reduces second
// request latency.
lazy_static! {
    pub static ref STEAMID_CACHE: Cache<u64, Player> = {
        info!("Initializing cache");
        let pool = CONFIGURATION.core.cache_thread_pool.unwrap_or(false);
        Cache::builder()
            .time_to_live(Duration::from_secs(60 * 60))
            .max_capacity(256)
            .thread_pool_enabled(pool)
            .build()
    };
}

lazy_static! {
    static ref CLIENT: Client = {
        Client::builder()
            .connect_timeout(Duration::from_millis(600))
            .timeout(Duration::from_millis(1000))
            .build()
            .expect("Trying to initalize Reqwest client for SteamAPI")
    };
}

#[derive(Deserialize)]
struct InternalPlayer {
    personaname: String,
    avatarhash: String,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Player {
    pub steamid: u64,
    pub name: String,
    pub avatar_url: String,
}

impl Default for Player {
    fn default() -> Self {
        Self {
            steamid: 0,
            name: "Player".into(),
            avatar_url: format!("{AVATAR_URL}/{DEFAULT_HASH}{POSTFIX}"),
        }
    }
}

#[tracing::instrument(level = "info")]
pub async fn get_player(steamid: u64) -> Result<Player, eyre::Error> {
    // Check if we have this SteamID in cache already.
    match STEAMID_CACHE.get(&steamid) {
        Some(v) => return Ok(v),
        None => {}
    }
    // Otherwise, bug the Steam API and deal with JSON.
    let client = (*CLIENT).clone();
    // Should really be using a request builder here instead of url formatting.
    let data = client
        .get(format!(
            "{API_URL}/?key={}&steamids={steamid}",
            CONFIGURATION.api.steam_key
        ))
        .send()
        .await?
        .text()
        .await?;
    let ply = deserialize_into_player(data, steamid)?;
    STEAMID_CACHE.insert(steamid, ply.clone());
    Ok(ply)
}

#[tracing::instrument(level = "debug")]
fn deserialize_into_player(data: String, steamid: u64) -> Result<Player, eyre::Error> {
    debug!("Fetching SteamID {steamid} from API");
    let mut untyped: Value = serde_json::from_str(&data)?;
    let player = untyped
        .get_mut("response")
        .and_then(|v| v.get_mut("players"))
        .and_then(|v| v.get_mut(0))
        .ok_or_else(|| eyre::Report::msg("Can't deserialize Response"))?;
    let ip: InternalPlayer = serde_json::from_value(player.take())?;
    Ok(Player {
        steamid,
        name: ip.personaname,
        avatar_url: format!("{AVATAR_URL}/{}{POSTFIX}", ip.avatarhash),
    })
}
